# -*- coding: utf-8 -*-
"""
TODO:
    * Add a list option to only list defined webfont in settings;
"""
import json, os, shutil, tempfile, zipfile

from django.conf import settings
from django.core.management.base import CommandError, BaseCommand
from django.template.loader import render_to_string

from icomoon.store import WebfontStore
from icomoon.utils import IcomoonSettingsError, extend_webfont_settings


class Command(BaseCommand):
    help = "Icomoon CLI to deploy an Icomoon webfont from downloaded ZIP archive."


    def add_arguments(self, parser):
        parser.add_argument('webfont_name', nargs='?', default='Default', help="Webfont key name from 'settings.ICOMOON_MANIFEST_FILEPATH'. (default: %(default)s)")
        parser.add_argument('archive_path', nargs='?', default='icomoon.zip', help="ZIP archive from icomoon to deploy. (default: %(default)s)")

    def handle(self, *args, **options):
        self.webfont_name = options['webfont_name']
        self.archive_path = options['archive_path']
        self.css_templatepath = options.get('css_template', settings.ICOMOON_CSS_TEMPLATE)
        self.requirements = options.get('requirements', settings.ICOMOON_ARCHIVE_REQUIREMENT)
        self.verbosity = int(options.get('verbosity'))

        self._info("* Using webfont name: {}", self.webfont_name)

        # Get the current manifest for given webfont name
        if self.webfont_name not in settings.ICOMOON_WEBFONTS:
            choices = "', '".join(settings.ICOMOON_WEBFONTS.keys())
            self._error("There is no defined webfont name '{name}' in your settings, choices are: '{choices}'", name=self.webfont_name, choices=choices)

        try:
            self.webfont_settings = extend_webfont_settings(settings.ICOMOON_WEBFONTS[self.webfont_name])
        except IcomoonSettingsError as e:
            self._error("Invalid webfont settings for '{}': {}", self.webfont_name, e.value)

        self._info("* Using webfont font dir: {}", self.webfont_settings['fontdir_path'])

        self.deploy()
        self._info("Finished !")

    def print_output(self, msg, *args, **kwargs):
        self.stdout.write(msg.format(*args, **kwargs))

    def _error(self, msg, *args, **kwargs):
        raise CommandError(msg.format(*args, **kwargs))

    def _debug(self, msg, *args, **kwargs):
        if self.verbosity >= 3:
            self.print_output(msg, *args, **kwargs)

    def _warning(self, msg, *args, **kwargs):
        if self.verbosity >= 2:
            self.print_output(msg, *args, **kwargs)

    def _info(self, msg, *args, **kwargs):
        if self.verbosity >= 1:
            self.print_output(msg, *args, **kwargs)

    def deploy(self):
        """
        Open a ZIP archive, validate requirements then deploy the webfont into
        project static files
        """
        self._info("* Opening archive: {}", self.archive_path)
        if not os.path.exists(self.archive_path):
            self._error("Given path does not exists: {}", self.archive_path)

        with zipfile.ZipFile(self.archive_path, 'r') as zip_archive:
            font_dir = self.requirements['font_dir']+'/'
            allowed_extensions = ['.'+item for item in self.requirements['extensions']]
            members = [member for member in zip_archive.namelist()]

            if settings.ICOMOON_MANIFEST_FILENAME not in members:
                raise self._error("Icomoon archive must contain a JSON manifest '{}'", settings.ICOMOON_MANIFEST_FILENAME)

            if font_dir not in members:
                raise self._error("Icomoon archive must contain the font directory '{}'", font_dir)

            # Scan for supported font files
            font_files = []
            for item in members:
                # Dont catch the font_dir itself nor sub directories, just files with allowed extensions
                if item.startswith(font_dir) and not item.endswith('/') and os.path.splitext(item)[-1] in allowed_extensions:
                    font_files.append(item)

            if not font_files:
                self._error("Font dir does not contain any supported format: {}", ', '.join(allowed_extensions))
            else:
                self._debug("* Finded font files in archive: {}", ', '.join(font_files))

            # Extract files from archive
            tmp_container, css_content = self.extract(zip_archive, font_files)

        # Install files
        self.install(tmp_container, font_dir, css_content)

    def extract(self, zip_archive, font_files):
        """
        Extract files to install
        """
        # Get a temp directory
        tmp_container = tempfile.mkdtemp(prefix='icomoon-tmp')
        self._debug("* Temporary dir for extracted archive: {}", tmp_container)

        # Extract manifest to temp directory
        zip_archive.extract(settings.ICOMOON_MANIFEST_FILENAME, tmp_container)
        # Then the font files
        for item in font_files:
            zip_archive.extract(item, tmp_container)

        # Get manifest for icon map
        webfont_store = WebfontStore(settings.ICOMOON_MANIFEST_FILENAME)
        webfont_store.get(self.webfont_name, {
            'fontdir_path': tmp_container,
        })
        icons = webfont_store.get_manifests()[self.webfont_name]
        #print json.dumps(icons, indent=4)

        # Render CSS icon part
        css_content = self.render_css(self.css_templatepath, icons)

        return tmp_container, css_content

    def render_css(self, template_path, icons):
        """
        Render CSS template to contain the icons stylesheet map
        """
        return render_to_string(template_path, {'icons': icons})

    def install(self, tmp_container, font_tmpdir, css_content):
        """
        Install extracted files and builded css
        """
        # Write builded css file to its destination
        with open(self.webfont_settings['csspart_path'], 'w') as css_file:
            css_file.write(css_content)

        # Clean previous font dir
        if os.path.exists(self.webfont_settings['fontdir_path']):
            shutil.rmtree(self.webfont_settings['fontdir_path'])
        font_srcdir = os.path.join(tmp_container, font_tmpdir)
        self._debug("* Installing font")
        self._debug("  - From: {}", font_srcdir)
        self._debug("  - To: {}", self.webfont_settings['fontdir_path'])

        shutil.copytree(font_srcdir, self.webfont_settings['fontdir_path'])

        # Copy new manifest into font dir
        manifest_src = os.path.join(tmp_container,
                                    settings.ICOMOON_MANIFEST_FILENAME)
        self._debug("* Installing manifest")
        self._debug("  - From: {}", manifest_src)
        self._debug("  - To: {}", self.webfont_settings['fontdir_path'])
        shutil.copy(manifest_src, self.webfont_settings['fontdir_path'])

        # Remove temp directory when all is done
        self._debug("* Removing temporary dir")
        shutil.rmtree(tmp_container)




"""
Icomoon manifest parser
"""
import io
import os
import json
from collections import OrderedDict

from six import string_types

from icomoon.utils import IcomoonSettingsError, extend_webfont_settings


class WebfontStore(object):
    """
    Webfont store to collect every given manifest.

    Store assume every webfont manifest file are located inside their webfont
    directory set in their ``fontdir_path`` option.

    Store does not create any file, it just parse manifest and store icon maps.

    Manifest filename is the same for every webfont, it is given as
    ``WebfontStore`` init argument ``manifest_filename``, commonly
    ``selection.json``.

    Args:
        manifest_filename (string): Manifest filename to search for in webfont
            directory.
    """
    def __init__(self, manifest_filename):
        self.manifest_filename = manifest_filename
        self.manifests = OrderedDict()
        self.errors = {}

    def get_icon_key(self, elem):
        """
        Return the icon name to be used as a sorting key
        """
        return elem.get('properties').get('name')

    def parse_manifest(self, fp):
        """
        Open manifest JSON file and build icon map

        Args:
            fp (string or fileobject): Either manifest filepath to open or
                manifest File object.

        Returns:
            dict: Webfont icon map. Contains:

            * ``class_name``: Builded icon classname with prefix configured
              in manifest (from parameters in Icomoon interface);
            * ``int``: Icon integer code like ``59649``;
            * ``hex``: Icon hexadecimal code like ``0xe901``;
            * ``unicode``: Icon unicode like ``U+E901``;
            * ``utf8``! Icon UTF8 code like ``\e901``;
        """
        # Given a string for file path to open
        if isinstance(fp, string_types):
            fp = io.open(fp, 'r', encoding='utf-8')

        with fp as json_file:
            webfont_manifest = json.load(json_file)

        # Get the font set prefix to know the css classname
        icon_prefix = webfont_manifest.get('preferences').get('fontPref').get('prefix')

        # Get sorted icons
        icons_map = OrderedDict()
        sorted_entries = sorted(webfont_manifest.get('icons'), key=self.get_icon_key)
        for icon_entry in sorted_entries:
            name = icon_entry.get('properties').get('name')
            code = icon_entry.get('properties').get('code')
            hexa_code = hex(code)

            icons_map[name] = {
                'class_name': icon_prefix + name,
                'int': code,
                'hex': hexa_code,
                'unicode': 'U+'+''.join(hex(code).split('x')[1:]).upper(),
                'utf8': '\\'+''.join(hex(code).split('x')[1:]).lower(),
            }

        return icons_map

    def get(self, webfont_name, webfont_settings):
        """
        Get a manifest file, parse and store it.

        Args:
            webfont_name (string): Webfont key name. Used to store manifest
                and potentially its parser error.
            webfont_settings (dict): Webfont settings (an item value from
                ``settings.ICOMOON_WEBFONTS``).
        """
        try:
            webfont_settings = extend_webfont_settings(webfont_settings)
        except IcomoonSettingsError as e:
            msg = "Invalid webfont settings for '{}': {}"
            self.errors[webfont_name] = msg.format(webfont_name, e.value)
            return

        filepath = os.path.join(webfont_settings['fontdir_path'],
                                self.manifest_filename)

        if os.path.exists(filepath):
            self.manifests[webfont_name] = self.parse_manifest(filepath)
        else:
            msg = ("""Filepath for webfont <strong>{name}</strong> does not """
                   """exists: <code>{filepath}</code>""")
            self.errors[webfont_name] = msg.format(name=webfont_name,
                                                   filepath=filepath)

    def fetch(self, webfonts):
        """
        Store every defined webfonts.

        Webfont are stored with sort on their name.

        Args:
            webfonts (dict): Dictionnary of webfont settings from
                ``settings.ICOMOON_WEBFONTS``.
        """
        sorted_keys = sorted(webfonts.keys())
        for webfont_name in sorted_keys:
            self.get(webfont_name, webfonts[webfont_name])

    def get_manifests(self):
        """
        Simple shortcut to get stored manifests.

        Returns:
            collections.OrderedDict: Stored webfont manifest.
        """
        return self.manifests



class IcomoonSettingsError(Exception):
    """Exception to raise when settings are invalid"""
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


def extend_webfont_settings(webfont_settings):
    """
    Validate a webfont settings and optionally fill missing ``csspart_path``
    option.

    Args:
        webfont_settings (dict): Webfont settings (an item value from
            ``settings.ICOMOON_WEBFONTS``).

    Returns:
        dict: Webfont settings
    """
    if not webfont_settings.get('fontdir_path', False):
        raise IcomoonSettingsError(("Webfont settings miss the required key "
                                    "item 'fontdir_path'"))

    if not webfont_settings.get('csspart_path', False):
        webfont_settings['csspart_path'] = None

    return webfont_settings



from setuptools import setup, find_packages

setup(
    name='django-icomoon',
    version=__import__('icomoon').__version__,
    description=__import__('icomoon').__doc__,
    long_description=open('README.rst').read(),
    author='David Thenon',
    author_email='sveetch@gmail.com',
    url='https://github.com/sveetch/django-icomoon',
    license='MIT',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        "Programming Language :: Python",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Framework :: Django',
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.10",
        "Framework :: Django :: 1.11",
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'six',
        'Django>=1.9,<1.12',
        'django-braces>=1.2.0',
    ],
    include_package_data=True,
    zip_safe=False
)